/*** DeHaze.cpp ***/

#include "MixedDomainFilter.h"
#include <cv.h>
#pragma comment(lib, "cv.lib")
#include <cxcore.h>
#pragma comment(lib, "cxcore.lib")
#include <highgui.h>
#pragma comment(lib, "highgui.lib")
#include <iostream>
using std::min;
using std::max;

double rat  = 1.00;
double ratE = 1.00;
double ratL = 1.00;

void getMinRGB(IplImage *orig, IplImage *minRGB, int W, int H, int C) {
	for(int i = 0;  i < W;  i++) {
		for(int j = 0;  j < H;  j++) {
			int tmp = int(unsigned char(orig->imageData[j*orig->widthStep+C*i]));
			for(int k = 1;  k < C;  k++) {
				tmp = min(tmp, int(unsigned char(orig->imageData[j*orig->widthStep+C*i+k])));
			}
			minRGB->imageData[j*minRGB->widthStep+i] = char(tmp);
		}
	}
}
void getDarkCh(IplImage *in, IplImage *ou, int W, int H, int len) {
	int *p = new int [max(W,H)+len*2];
	int *q = new int [max(W,H)+len*2];
	for(int i = 0;  i < W;  i++) {
		int s = 0;
		int t = 0;
		for(int j = 0;  j < H+len;  j++) {
			if(j < H) {
				int tmp = int(unsigned char(in->imageData[j*in->widthStep+i]));
				while(t > s && q[t-1] >= tmp) {
					t--;
				}
				q[t] = tmp;
				p[t++] = j;
			}
			if(j >= len) {
				ou->imageData[(j-len)*ou->widthStep+i] = q[s];
			}
			if(p[s]+len*2 == j) {
				s++;
			}
		}
	}
	for(int j = 0;  j < H;  j++) {
		int s = 0;
		int t = 0;
		for(int i = 0;  i < W+len;  i++) {
			if(i < W) {
				int tmp = int(unsigned char(ou->imageData[j*ou->widthStep+i]));
				while(t > s && q[t-1] >= tmp) {
					t--;
				}
				q[t] = tmp;
				p[t++] = i;
			}
			if(i >= len) {
				ou->imageData[j*ou->widthStep+i-len] = q[s];
			}
			if(p[s]+len*2 == i) {
				s++;
			}
		}
	}
	for(int i = 0;  i < W;  i++) {
		for(int j = 0;  j < H;  j++) {
			ou->imageData[j*ou->widthStep+i] = unsigned char(min(255.0, ratE*unsigned char(ou->imageData[j*ou->widthStep+i])));
		}
	}
	delete []p;
	delete []q;
}
void getAtmosphere(IplImage *orig, IplImage *ch, int atm[], int W, int H, int C, int len) {
	int tot = 0;
	int best = 0;
	for(int i = 0;  i < W;  i++) {
		for(int j = 0;  j < H;  j++) {
			if(i < len*5 || i+len*5 >= W || j < len*5 || j+len*5 >= H) {
				if(unsigned char(ch->imageData[j*ch->widthStep+i]) > best) {
					best = unsigned char(ch->imageData[j*ch->widthStep+i]);
				}
			}
		}
	}
	for(int i = 0;  i < W;  i++) {
		for(int j = 0;  j < H;  j++) {
			if(i < len*5 || i+len*5 >= W || j < len*5 || j+len*5 >= H) {
				if(unsigned char(ch->imageData[j*ch->widthStep+i]) > best-20) {
					for(int k = 0;  k < C;  k++) {
						atm[k] += unsigned char(orig->imageData[j*orig->widthStep+i*C+k]);
					}
					tot++;
				}
			}
		}
	}
	for(int k = 0;  k < C;  k++) {
		atm[k] /= tot;
	}
}
inline void convert2Dto1D(IplImage *im, double *ar, int W, int H) {
	for(int i = 0;  i < W;  i++) {
		for(int j = 0;  j < H;  j++) {
			ar[j*W+i] = unsigned char(im->imageData[j*im->widthStep+i])/255.0;
		}
	}
}
inline void convert1Dto2D(double *ar, IplImage *im, int W, int H) {
	for(int i = 0;  i < W;  i++) {
		for(int j = 0;  j < H;  j++) {
			im->imageData[j*im->widthStep+i] = unsigned char(ar[j*W+i]*255+0.50);
		}
	}
}
void hazeRemove(IplImage *orig, IplImage *res, double *deep, int W, int H, int C, int atm[]) {
	for(int i = 0;  i < W;  i++) {
		for(int j = 0;  j < H;  j++) {
			double tmp = min(0.98, deep[j*W+i]);
			for(int k = 0;  k < C;  k++) {
				res->imageData[j*res->widthStep+C*i+k] = unsigned char(0.4+min(255.0, max(0.0,
				ratL*(unsigned char(orig->imageData[j*orig->widthStep+C*i+k])-tmp*atm[k])/(1.00-tmp))));
			}
		}
	}
}
void deHaze(string xFileName, string yFileName, double Rat, double RatE, double RatL, int level) {
	rat  = Rat;
	ratE = RatE;
	ratL = RatL;
	IplImage* orig = cvLoadImage(xFileName.c_str(), -1);
	int W = orig->width;
	int H = orig->height;
	int C = orig->nChannels;
	int len = int(min(W, H)/200.0+0.50);

	IplImage *yImageFloat = cvCreateImage(cvGetSize(orig), 32, 3);
	IplImage *minRGB = cvCreateImage(cvGetSize(orig), IPL_DEPTH_8U, 1);
	getMinRGB(orig, minRGB, W, H, C);

	IplImage *darkCh = cvCreateImage(cvGetSize(orig), IPL_DEPTH_8U, 1);
	getDarkCh(minRGB, darkCh, W, H, len);
	
	int atm[3];
	memset(atm, 0, sizeof(atm));
	getAtmosphere(orig, darkCh, atm, W, H, C, len);
	for(int k = 0;  k < 3;  k++) {
		atm[k] = int(atm[k]*rat+0.50);
	}

	double *x1 = new double[W*H];
	double *x2 = new double[W*H];
	convert2Dto1D(minRGB, x1, W, H);
	convert2Dto1D(darkCh, x2, W, H);
	double *haDeep = new double[W*H];

	cvCvtScale(orig, yImageFloat, 1.0/255);
	IplImage *yImageLabFloat = cvCloneImage(yImageFloat);
	cvCvtColor(yImageFloat, yImageLabFloat, CV_BGR2Lab);

	double **lab = new double *[3];
	for(int i = 0;  i < 3;  i++) {
		lab[i] = new double[W*H];
	}
	for(int i = 0;  i < W;  i++) {
		for(int j = 0;  j < H;  j++) {
			for(int k = 0;  k < 3;  k++) 
				lab[k][j*W+i] = *((float *)(&yImageLabFloat->imageData[j*yImageLabFloat->widthStep+12*i+4*k]))/100;
		}
	}
	cvReleaseImage(&yImageFloat);
	cvReleaseImage(&yImageLabFloat);
	jointPyramidFilter(W, H, x1, lab, haDeep, 1.00, 0.00, 1.00, 0.20, false, false, level, 1.00);

	IplImage *deep = cvCreateImage(cvGetSize(orig), IPL_DEPTH_8U, 1);
	convert1Dto2D(haDeep, deep, W, H);

	IplImage *result = cvCloneImage(orig);
	hazeRemove(orig, result, haDeep, W, H, C, atm);

//	cvSaveImage((yFileName+"2.png").c_str(), minRGB);
//	cvSaveImage((yFileName+"4.png").c_str(), darkCh);
//	cvSaveImage((yFileName+"3.png").c_str(), deep);
	cvSaveImage(yFileName.c_str(), result);

	cvReleaseImage(&orig);
	cvReleaseImage(&minRGB);
	cvReleaseImage(&darkCh);
	cvReleaseImage(&deep);
	cvReleaseImage(&result);

	delete []x1;
	delete []x2;
	delete []haDeep;
	for(int i = 0;  i < 3;  i++) {
		delete []lab[i];
	}
	delete []lab;
}
